#ifndef READWTB_DISASSEMBLER_H
#define READWTB_DISASSEMBLER_H

#include "welltype/wtruntime.h"
#include "welltype/byteorder.h"
#include "wtruntime/types.h"

struct Disassembler;
struct labelmap;


#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

struct Disassembler* Disassembler_Create(void);
void Disassembler_Release(struct Disassembler* const disasm);


/* Disassemble the given memory. */
/* Returns NULL on success; otherwise the address of the unknown instruction. */
/* Microprogram references will be resolved to strings only when 'vermatch' is true. */
/* 'addr' is the logical address of the code (in the address space of the program). */
/* 'size' is the code size in octets. 'labels' is optional, and maps program-space */
/* addresses to labels (debug labels, and function names). 'bo' is the byte order */
/* ID read from the WTB. */
const wt_u16* Disassembler_Disassemble(struct Disassembler* disasm, wt_bool vermatch,
	wt_u32 addr, const wt_u16* code, wt_u32 size, struct labelmap* labels, wt_bo bo);


#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* READWTB_DISASSEMBLER_H */
