/* $Id: sym.h 7234 2024-01-11 19:57:50Z aron $ */

#ifndef WELLTYPE_SYM_H
#define WELLTYPE_SYM_H

#include "config.h"
#include "wtruntime.h"

#define WTR_SYM_UNDEF           0xFFFFFFFFu /* look up for symbol definition */

#define WTR_SYM_OP_ASSOC_LEFT   1
#define WTR_SYM_OP_ASSOC_RIGHT  2
#define WTR_SYM_OP_PREC_MIN     1
#define WTR_SYM_OP_PREC_MAX     15

struct wtr_symtab;

enum wtr_symtype
{
	WTR_SYMTYPE_WTFUNC = 1,
	WTR_SYMTYPE_EXC,
	WTR_SYMTYPE_OPER,

#define WTR_SYMTYPE__FIRST_TYPE WTR_SYMTYPE_TYPEAL
	WTR_SYMTYPE_TYPEAL,
	WTR_SYMTYPE_SUBTYPE,
	WTR_SYMTYPE_DATATYPE,
	WTR_SYMTYPE_REC,
	WTR_SYMTYPE_PRIVREC,
	WTR_SYMTYPE_LIMREC,
	WTR_SYMTYPE_FUNCTYPE,
	WTR_SYMTYPE_ENUM,
#define WTR_SYMTYPE__LAST_TYPE WTR_SYMTYPE_ENUM

	WTR_SYMTYPE_CFUNC = 0x100
};

#define wtr_sym_is_exc(_symtype_) \
	(WTR_SYMTYPE_EXC==(_symtype_))

#define wtr_sym_is_oper(_symtype_) \
	(WTR_SYMTYPE_OPER==(_symtype_))

#define wtr_sym_is_type(_symtype_) \
	(WTR_SYMTYPE__FIRST_TYPE<=(_symtype_) && \
	 WTR_SYMTYPE__LAST_TYPE>=(_symtype_))

#define wtr_sym_is_func(_symtype_) \
	(WTR_SYMTYPE_WTFUNC==(_symtype_) || WTR_SYMTYPE_CFUNC==(_symtype_))

enum wtr_symbind
{
	WTR_SYMBIND_IMPORT = 1,
	WTR_SYMBIND_EXPORT
};

struct wtr_sym
{
	wt_uint16_t        type;   /* enum wtr_symtype */
	wt_uint8_t         bind;   /* enum wtr_symbind */
	struct wtr_symtab* symtab; /* owner symtab */
	const char*        name;   /* symbol name */
};

struct wtr_sym_exc
{
	struct wtr_sym sym;
	wt_exception   exc_id;  /* global exception identifier */
};

struct wtr_sym_oper
{
	struct wtr_sym sym;
	wt_globalid    oper_id;
	wt_uint8_t     precedence;
	wt_uint8_t     associativity;
};

struct wtr_sym_type
{
	struct wtr_sym sym;
	wt_typeid      typeinfo; /* runtime type info */
};

struct wtr_s_recfield
{
	const char* name;
	wt_typeid   type;
};

struct wtr_sym_rec
{
	struct wtr_sym_type    type;
	wt_uint32_t            fldc;
	struct wtr_s_recfield* fldv;
};

struct wtr_sym_privrec
{
	struct wtr_sym_type type;
};

struct wtr_sym_limrec
{
	struct wtr_sym_type type;
};

struct wtr_s_datacons
{
	const char* name;
	wt_uint32_t typc;
	wt_typeid*  typv;
};

struct wtr_sym_datatype
{
	struct wtr_sym_type    type;
	wt_uint32_t            numcons;
	wt_uint32_t            defcons;
	struct wtr_s_datacons* conses;
};

struct wtr_s_funcsig
{
	wt_uint32_t argc;
	wt_uint32_t retc;
	wt_typeid*  argv;
	wt_typeid*  retv;
	wt_uint8_t  is_pure;
};

struct wtr_sym_functype
{
	struct wtr_sym_type  type;
	struct wtr_s_funcsig sig;
};

struct wtr_sym_enum
{
	struct wtr_sym_type type;
	wt_uint32_t         enumeratorc;
	const char**        enumeratorv;
};

struct wtr_sym_func
{
	struct wtr_sym       sym;
	struct wtr_s_funcsig sig;
	wt_globalid          func_id;
};

struct wtr_sym_wtfunc
{
	struct wtr_sym_func func;
	wt_uint32_t         entry; /* function entry */
};

struct wtr_sym_cfunc
{
	struct wtr_sym_func func;  /* 'func.sig.is_pure' must be always 0 */
	WT_PFN_EXTFUNC      cfunc;
	void*               token;
};

#endif /* WELLTYPE_SYM_H */
