/* $Id: wtb4.h 6915 2023-11-14 20:41:53Z aron $ */

#ifndef WELLTYPE_WTB4_H
#define WELLTYPE_WTB4_H

#include "config.h"
#include <stddef.h>

typedef wt_uint8_t  wt_m16sndx_t;
typedef wt_uint8_t  wt_m16half_t;
typedef wt_uint16_t wt_m16addr_t;
typedef wt_uint16_t wt_m16offs_t;
typedef wt_uint16_t wt_m16size_t;

typedef wt_uint16_t wt_m32sndx_t;
typedef wt_uint16_t wt_m32half_t;
typedef wt_uint32_t wt_m32addr_t;
typedef wt_uint32_t wt_m32offs_t;
typedef wt_uint32_t wt_m32size_t;

#if WELLTYPE_CONF_HAS_INT64

typedef wt_uint32_t wt_m64sndx_t;
typedef wt_uint32_t wt_m64half_t;
typedef wt_uint64_t wt_m64addr_t;
typedef wt_uint64_t wt_m64offs_t;
typedef wt_uint64_t wt_m64size_t;

#endif /* WELLTYPE_CONF_HAS_INT64 */

#define WTB4MAG0               'W'
#define WTB4MAG1               'T'
#define WTB4MAG2               'B'
#define WTB4MAG3               '\x04'

#define WTB4V_CURRENT          1

#define WTB4BO_LITTLE          1 /* little endian */
#define WTB4BO_BIG             2 /* big endian */

#define WTB4M_TOMB             0x30 /* 32-bit Tomb */
#define WTB4M_TOMB2_16         0x41 /* 16-bit Tomb2 */
#define WTB4M_TOMB2_32         0x42 /* 32-bit Tomb2 */
#define WTB4M_TOMB2_64         0x43 /* 64-bit Tomb2 */

#define WTB4DATA_16            1  /* 16-bit hdr */
#define WTB4DATA_32            2  /* 32-bit hdr */
#define WTB4DATA_64            3  /* 64-bit hdr */

#define WTB4T_REL              1 /* relocatable: collection of functions */
#define WTB4T_DYN              2 /* dynamic executable: linked and ready to be loaded */
#define WTB4T_EXEC             3 /* flat executable: linked against an API and ready to be executed */
#define WTB4T_API              4 /* API to link flat executables against */

#define WTB4_NIDENT            16 /* size of 'b_ident' */

#define WTB4I_MAGIC0           0  /* value is WTB4MAG0 */
#define WTB4I_MAGIC1           1  /* value is WTB4MAG1 */
#define WTB4I_MAGIC2           2  /* value is WTB4MAG2 */
#define WTB4I_MAGIC3           3  /* value is WTB4MAG3 */
#define WTB4I_VERSION          4  /* WTB4V_CURRENT */
#define WTB4I_BYTEORDER        5  /* byte order (WTB4BO_...) */
#define WTB4I_DATAMODEL        6  /* data model (WTB4DATA_...) */
#define WTB4I_MACHINE          7  /* machine/instruction set (WTB4M_...) */
#define WTB4I_ABIVER           8  /* minimal required ABI version */
#define WTB4I_TYPE             9  /* binary type (WTB4T_...) */
#define WTB4I_FEATURES1        10 /* primary machine features */
/*#define WTB4I_               11 */ /* MBZ */
#define WTB4I_SHENTSIZE        12 /* section header entry size */
#define WTB4I_COMPVERMAJOR     13 /* compiler version (major) */
#define WTB4I_COMPVERMINOR     14 /* compiler version (minor) */
#define WTB4I_COMPVERPATCH     15 /* compiler version (patch) */

struct wtb4_hdr16
{
	wt_uint8_t   b_ident[WTB4_NIDENT]; /* identification (item indices are WTB4I_*) */
	wt_m16offs_t b_shoff;              /* absolute offset to section header */
	wt_m16sndx_t b_shnum;              /* number of sections */
	wt_m16sndx_t b_shstrndx;           /* section header string table section index */
};

struct wtb4_hdr32
{
	wt_uint8_t   b_ident[WTB4_NIDENT]; /* identification (item indices are WTB4I_*) */
	wt_m32offs_t b_shoff;              /* absolute offset to section header */
	wt_m32sndx_t b_shnum;              /* number of sections */
	wt_m32sndx_t b_shstrndx;           /* section header string table section index */
};

#if WELLTYPE_CONF_HAS_INT64

struct wtb4_hdr64
{
	wt_uint8_t   b_ident[WTB4_NIDENT]; /* identification (item indices are WTB4I_*) */
	wt_m64offs_t b_shoff;              /* absolute offset to section header */
	wt_m64sndx_t b_shnum;              /* number of sections */
	wt_m64sndx_t b_shstrndx;           /* section header string table section index */
};

#endif /* WELLTYPE_CONF_HAS_INT64 */

#define WTB4ST_NULL      0x00 /* unused entry */
#define WTB4ST_NOTE      0x01 /* note -- not directly related to the program binary */
#define WTB4ST_PROGBITS  0x02 /* program binary, uninterpretable by its own */
#define WTB4ST_STRTAB    0x03 /* string table */
#define WTB4ST_SYMTAB    0x04 /* symbol table, entry type is wtb4_symXX */
#define WTB4ST_PBSWAP    0x05 /* progbits byte swap info */
#define WTB4ST_DYNAMIC   0x06 /* dynamic loader info, entry type is wtb_dynXX */
#define WTB4ST_INTEGR    0x07 /* integrity (wtb4_integr; wt_uint8_t[]) */
#define WTB4ST_RELOC     0x08 /* relocation info */
#define WTB4ST_METADATA  0x09 /* metadata (contains key-value zero-terminated string pairs) */
#define WTB4ST_API       0x0A /* TODO */
#define WTB4ST_DEBUGMAP  0x70 /* debug: map offsets with labels, entry type is wtb_debugmapXX */

#define WTB4SF_READ      0x01 /* content needs to be readable after load */
#define WTB4SF_WRITE     0x02 /* content needs to be writable after load */
#define WTB4SF_EXEC      0x04 /* contains executable code */

struct wtb4_shdr16
{
	wt_m16half_t sh_type;
	wt_m16half_t sh_flags;
	wt_m16offs_t sh_offs;
	wt_m16size_t sh_fsize;
	wt_m16size_t sh_msize;
	wt_m16sndx_t sh_link;

	/* WTBT_REL-specific fields */

	wt_m16half_t sh_entsz;
	wt_m16offs_t sh_name;
};

#define WTB4DT_DATABSS   0x01 /* data+bss segment */
#define WTB4DT_PLT       0x02 /* program linkage table */
#define WTB4DT_SYMTAB    0x03 /* the symbol table to use (dyn_value is num syms) */
#define WTB4DT_ENTRY     0x04 /* section-relative PC of the main function */
#define WTB4DT_CTOR      0x05 /* section-relative PC of the constructor */
#define WTB4DT_DTOR      0x06 /* section-relative PC of the destructor */
#define WTB4DT_STKSZ     0x07 /* stack size advice */
#define WTB4DT_LIB       0x08 /* dynamic library to load (dyn_value is a name in an strtab) */
#define WTB4DT_REL       0x09 /* relocation info */

struct wtb4_dyn16
{
	wt_m16half_t dyn_type;  /* dynamic attribute type (WTB4DT_...) */
	wt_m16sndx_t dyn_sndx;  /* section reference */
	wt_m16addr_t dyn_value; /* value of the attribute */
};

#define WTB4SYMT_NOTYPE    0x00 /* purpose of the symbol is not defined (bind must be LOCAL) */
#define WTB4SYMT_FUNC      0x01 /* function */
#define WTB4SYMT_EXC       0x02 /* exception */
#define WTB4SYMT_OPER      0x03 /* operator */
#define WTB4SYMT_TYPE      0x04 /* type alias */
#define WTB4SYMT_SUBTYPE   0x05 /* sub-type */
#define WTB4SYMT_DATATYPE  0x06 /* data type */
#define WTB4SYMT_REC       0x07 /* record */
#define WTB4SYMT_PRIVREC   0x08 /* private record */
#define WTB4SYMT_LIMREC    0x09 /* limited record */
#define WTB4SYMT_FUNCTYPE  0x0A /* function type */
#define WTB4SYMT_ENUM      0x0B /* enum */

#define WTB4SYMB_LOCAL     0x0 /* internal symbol, does not affect dynamic linker */
#define WTB4SYMB_GLOBAL    0x1 /* instructs the dynamic linker to check and fill out the missing information */

#define WTB4SYM_TYPE(info)        ((wt_uint8_t)((info) & 0x3F))
#define WTB4SYM_BIND(info)        ((wt_uint8_t)((info) >> 6))
#define WTB4SYM_INFO(type, bind)  (((bind) << 6) | ((type) & 0x3F))

struct wtb4_sym16
{
	wt_m16offs_t sym_data; /* offset to symbol data in current section; symbol table ends with 0 */
	wt_m16sndx_t sym_sndx; /* section reference */
	wt_uint8_t   sym_info; /* symbol type and bind (WTB4SYMT_..., WTB4SYMB_...) */
};

struct wtb4_sym32
{
	wt_m32offs_t sym_data; /* offset to symbol data in current section; symbol table ends with 0 */
	wt_m32sndx_t sym_sndx; /* section reference */
	wt_uint8_t   sym_info; /* symbol type and bind (WTB4SYMT_..., WTB4SYMB_...) */
	wt_uint8_t   sym_pad[1]; /* padding, must be zero */
};

struct wtb4_sym64
{
	wt_m64offs_t sym_data; /* offset to symbol data in current section; symbol table ends with 0 */
	wt_m64sndx_t sym_sndx; /* section reference */
	wt_uint8_t   sym_info; /* symbol type and bind (WTB4SYMT_..., WTB4SYMB_...) */
	wt_uint8_t   sym_pad[3]; /* padding, must be zero */
};

/* predefined (or basic) types */

#define WTB4TYPE_BOOL           1
#define WTB4TYPE_I8             2
#define WTB4TYPE_U8             3
#define WTB4TYPE_I16            4
#define WTB4TYPE_U16            5
#define WTB4TYPE_I32            6
#define WTB4TYPE_U32            7
#define WTB4TYPE_I64            8
#define WTB4TYPE_U64            9
#define WTB4TYPE_F32            10
#define WTB4TYPE_F64            11
#define WTB4TYPE_CHAR           12
#define WTB4TYPE_STRING         13
#define WTB4TYPE_HANDLE         14
#define WTB4TYPE_FILE           15
#define WTB4TYPE_DIR            16

/* flags for type info */

#define WTB4TF_REF              0x1
#define WTB4TF_SEQ              0x2
#define WTB4TF_MUTABLE          0x4 /* function argument type modifier: able to modify object content */
#define WTB4TF_LOCAL            0x8 /* type is defined in the symtab */

/* type info composition/decomposition; type info is compacted by the */
/* algorithm as seen in 2.2.3 "Remaining Length" in MQTT specification */
/* v3.1.1, but for up to 10 bytes (on 64-bit machines) */

#define WTB4TINFO(flags,type)   (((type)<<4) | ((flags) & 0xF))
#define WTB4TINFO_TYPE(tinfo)   ((tinfo) >> 4)
#define WTB4TINFO_FLAGS(tinfo)  ((tinfo) & 0xF)

#define WTB4OPINFO(prec,assoc)  (((assoc)<<6) | ((prec) & 0x3F))
#define WTB4OPINFO_ASSOC(info)  ((info) >> 6)
#define WTB4OPINFO_PREC(info)   ((info) & 0x3F)
#define WTB4OPINFO_ASSOC_LEFT   1
#define WTB4OPINFO_ASSOC_RIGHT  2

/*
  Symbol data specification:

  WTB4SYMT_FUNCTYPE:
    name (zero-term string)
    pure spec (uint8, 0 or 1)
    arg count (compacted machine word)
    retval count (compacted machine word)
    [
      arg type info (compacted machine word)
    ]
    [
      retval type info (compacted machine word)
    ]

  WTB4SYMT_FUNC:
    offset of the function in sym_sndx (machine word, unaligned, 0xFF..FF is unknown)
    *rest is the same as WTB4SYMT_FUNCTYPE

  WTB4SYMT_EXC:
    name (zero-term string)

  WTB4SYMT_TYPE:
  WTB4SYMT_SUBTYPE:
    name (zero-term string)
    type (compacted machine word)

  WTB4SYMT_DATATYPE:
    name [zero-term string]
    consnum [compacted machine word]
    defcons [compacted machine word]
    [
      cons name (zero-term string)
      type count (compacted machine word)
      [
        type info (compacted machine word)
      ]
    ]

  WTB4SYMT_OPER:
    name (zero-term string)
    opinfo (uint8, WTB4OPINFO)

  WTB4SYMT_REC:
    name (zero-term string)
    field count (compacted machine word)
    [
      field type info (compacted machine word)
      field name (zero-term string)
    ]

  WTB4SYMT_PRIVREC (unique):
    name (zero-term string)

  WTB4SYMT_LIMREC (unique):
    name (zero-term string)

  WTB4SYMT_ENUM:
    name (zero-term string)
    enumerator count (compacted machine word)
    [
      enumerator name (zero-term string)
    ]

*/

#define WTB4PBS_NOSWAP  1
#define WTB4PBS_SWAP16  2
#define WTB4PBS_SWAP32  4
#define WTB4PBS_SWAP64  8

/*
	ProgBits Swap section content:
	{[swap size (WTB4PBS_...)] [ent count (compacted machine word)]}*
*/

struct wtb4_integr
{
	wt_uint16_t hdrsz; /* size of the header */
	wt_uint16_t algo;  /* hash algoritm (WTBHA_...) */
};

#define WTB4REL_FUNC        1
#define WTB4REL_EXC         2
#define WTB4REL_CALL        3 /* WTB4T_REL only */
#define WTB4REL_DATA        4 /* WTB4T_REL only */

struct wtb4_rel16
{
	wt_m16half_t r_type; /* relocation type (WTB4REL_...) */
	wt_m16sndx_t r_sndx; /* perform relocation in this section */
	wt_m16offs_t r_offs; /* offset of the relocation in the section */
	wt_m16offs_t r_sym;  /* symbol index in symtab (sh_link) */
};

/* ends with dm_offs==0 && dm_name==0 */
struct wtb4_debugmap16
{
	wt_m16offs_t dm_offs; /* offset in sh_link */
	wt_m16offs_t dm_name; /* offset to name in current section (zero-terminated string) */
};

struct wt_fptr16
{
	wt_m16addr_t prog;
	wt_m16addr_t pc;
};

struct wt_fptr32
{
	wt_m32addr_t prog;
	wt_m32addr_t pc;
};

#if WELLTYPE_CONF_HAS_INT64

struct wt_fptr64
{
	wt_m64addr_t prog;
	wt_m64addr_t pc;
};

#endif /* WELLTYPE_CONF_HAS_INT64 */

#endif /* WELLTYPE_WTB4_H */
