/* $Id: ast_functiontype.h 6310 2023-01-16 20:39:48Z aron $ */
/*
    Welltype parser and AST
    Copyright (C) 2014-2023  Aron Barath

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#ifndef WTPARSER_AST_FUNCTIONTYPE_H
#define WTPARSER_AST_FUNCTIONTYPE_H

#include "../astnode.h"
#include "ast_type.h"

namespace wtast
{

class FunctionSignature;

class FunctionType : public Type
{
	FunctionSignature* mFuncSig;
	wt_u32             mTypeId;

public:
	FunctionType(FunctionSignature* funcsig, wt_u32 type_id)
		: Type()
		, mFuncSig(funcsig)
		, mTypeId(type_id)
	{
	}

	~FunctionType();

	void Accept(AstVisitor* vis);

	FunctionType* Copy();

	Kind GetKind() const
	{
		return FUNCTIONTYPE;
	}

	FunctionSignature* GetFunctionSignature()
	{
		return mFuncSig;
	}

	std::string GetName() const;

	wt_u32 GetTypeId() const
	{
		return mTypeId;
	}

	wt_u32 GetLoadSize() const;

	bool IsObject() const
	{
		return false;
	}

	bool IsDefaultConstructible() const
	{
		return false;
	}

	bool IsFunctionType() const
	{
		return true;
	}

	bool HasFunctionType() const
	{
		return true;
	}
};

} // namespace wtast

#endif // WTPARSER_AST_FUNCTIONTYPE_H
