/* $Id: ast_seqtype.h 7194 2024-01-08 19:29:54Z aron $ */
/*
    Welltype parser and AST
    Copyright (C) 2014-2023  Aron Barath

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#ifndef WTPARSER_AST_SEQTYPE_H
#define WTPARSER_AST_SEQTYPE_H

#include "../astnode.h"
#include "ast_type.h"
#include "welltype/wtb3.h"

namespace wtast
{

class SeqType : public Type
{
	Type*       mType;
	wt_u32      mTypeId;
	std::string mName;

	void GenerateName();

public:
	SeqType(Type* type, wt_u32 type_id)
		: Type()
		, mType(type)
		, mTypeId(type_id)
	{
		GenerateName();
	}

	~SeqType();

	void Accept(AstVisitor* vis);

	SeqType* Copy();

	Kind GetKind() const
	{
		return SEQTYPE;
	}

	Type* GetElementType()
	{
		return mType;
	}

	std::string GetName() const
	{
		return mName;
	}

	wt_u32 GetTypeId() const
	{
		return mTypeId;
	}

	wt_u32 GetLoadSize() const;

	bool IsObject() const
	{
		return true;
	}

	bool IsSequence() const
	{
		return true;
	}

	bool HasFunctionType() const
	{
		return mType->HasFunctionType();
	}

	bool IsCopyable() const
	{
		return mType->IsCopyable();
	}

	void SetMutable()
	{
		mTypeId |= WTB3TF_MUTABLE;
	}
};

} // namespace wtast

#endif // WTPARSER_AST_SEQTYPE_H
