//================================================================================================
// $Id: frontend.h 6309 2023-01-16 20:19:42Z aron $
//
// Project:
//		Welltype Parser
// Contents:
//		Parser frontend
// Authors:
//		Aron Barath, 2014-2023
//================================================================================================
// Copyright (c) Aron Barath
//================================================================================================

#ifndef WTPARSER_FRONTEND_H
#define WTPARSER_FRONTEND_H

#include <string>

namespace wtast
{
class Program;
} // wtast

namespace wtparser
{

class IFile
{
public:
	IFile() { }
	virtual ~IFile() { }
	virtual bool Eof() const = 0;
	virtual unsigned int Read(void* buffer, unsigned int size) = 0;
};

class ISourceManager
{
public:
	ISourceManager() { }
	virtual ~ISourceManager() { }
	virtual IFile* OpenFile(const std::string & filename) = 0;
};

class IProblemHandler
{
public:
	IProblemHandler() { }
	virtual ~IProblemHandler() { }
	virtual void AddNote(const std::string & /*file*/, unsigned int /*line*/, const std::string & /*msg*/) { }
	virtual void AddWarning(const std::string & /*file*/, unsigned int /*line*/, const std::string & /*msg*/) { }
	virtual void AddError(const std::string & /*file*/, unsigned int /*line*/, const std::string & /*msg*/) { }
};

struct IOptions
{
	virtual ~IOptions() { }
	virtual bool DebugLexer() const { return true; } // TODO: false
	virtual bool DebugRules() const { return true; } // TODO: false
	virtual bool AllowFloatingBlocks() const { return true; }
};

wtast::Program* Parse(const std::string & filename, ISourceManager* sm, IProblemHandler* ph, const IOptions* opt);

} // namespace wtparser

#endif // WTPARSER_FRONTEND_H
