#include "config_private.h"
#include "byteorder.h"

#define __SWAP_16(_short_) ((wt_uint16_t)( \
		(((wt_uint16_t)(_short_))>>8) | \
		(((wt_uint16_t)(_short_))<<8)  \
	))

#define __SWAP_32(_int_) ((wt_uint32_t)( \
		(((wt_uint32_t)(_int_))<<24) | \
		((((wt_uint32_t)(_int_))<<8) & 0x00ff0000ul) | \
		((((wt_uint32_t)(_int_))>>8) & 0x0000ff00ul) | \
		(((wt_uint32_t)(_int_))>>24) \
	))

#define __SWAP_64(_long_) ((wt_uint64_t)( \
		(((wt_uint64_t)(_long_))<<(8*7)) | \
		((((wt_uint64_t)(_long_))<<(8*5)) & 0x00ff000000000000ull) | \
		((((wt_uint64_t)(_long_))<<(8*3)) & 0x0000ff0000000000ull) | \
		((((wt_uint64_t)(_long_))<<(8*1)) & 0x000000ff00000000ull) | \
		((((wt_uint64_t)(_long_))>>(8*1)) & 0x00000000ff000000ull) | \
		((((wt_uint64_t)(_long_))>>(8*3)) & 0x0000000000ff0000ull) | \
		((((wt_uint64_t)(_long_))>>(8*5)) & 0x000000000000ff00ull) | \
		(((wt_uint64_t)(_long_))>>(8*7)) \
	))

wt_bo wtr_byteorder(void)
{
	static union bint_t
	{
		wt_uint32_t i;
		wt_u8  b[4];
	} bint = { (wt_uint32_t)0x02030401u };

	return bint.b[0];
}

wt_uint16_t wtr_swap16(wt_uint16_t x)
{
	return __SWAP_16(x);
}

wt_uint32_t wtr_swap32(wt_uint32_t x)
{
	return __SWAP_32(x);
}

wt_uint64_t wtr_swap64(wt_uint64_t x)
{
	return __SWAP_64(x);
}

static wt_uint16_t wtr_id16(wt_uint16_t x)
{
	return x;
}

static wt_uint32_t wtr_id32(wt_uint32_t x)
{
	return x;
}

static wt_uint64_t wtr_id64(wt_uint64_t x)
{
	return x;
}

static const struct wtr_swaptab wtr_swaptab_swap =
{
	wtr_swap16, wtr_swap16,
	wtr_swap32, wtr_swap32,
	wtr_swap64, wtr_swap64
};

static const struct wtr_swaptab wtr_swaptab_id =
{
	wtr_id16, wtr_id16,
	wtr_id32, wtr_id32,
	wtr_id64, wtr_id64
};

const struct wtr_swaptab* wtr_swaptab(wt_bo bo)
{
	if(bo==wtr_byteorder())
	{
		return &wtr_swaptab_id;
	}

	return &wtr_swaptab_swap;
}

wt_uint16_t wtr_botoh16(wt_uint16_t x, wt_bo bo)
{
	if(bo==wtr_byteorder())
	{
		return x;
	}

	return __SWAP_16(x);
}

wt_uint32_t wtr_botoh32(wt_uint32_t x, wt_bo bo)
{
	if(bo==wtr_byteorder())
	{
		return x;
	}

	return __SWAP_32(x);
}

wt_uint64_t wtr_botoh64(wt_uint64_t x, wt_bo bo)
{
	if(bo==wtr_byteorder())
	{
		return x;
	}

	return __SWAP_64(x);
}

wt_uint16_t wtr_htobo16(wt_uint16_t x, wt_bo bo)
{
	if(bo==wtr_byteorder())
	{
		return x;
	}

	return __SWAP_16(x);
}

wt_uint32_t wtr_htobo32(wt_uint32_t x, wt_bo bo)
{
	if(bo==wtr_byteorder())
	{
		return x;
	}

	return __SWAP_32(x);
}

wt_uint64_t wtr_htobo64(wt_uint64_t x, wt_bo bo)
{
	if(bo==wtr_byteorder())
	{
		return x;
	}

	return __SWAP_64(x);
}
