#ifndef WTRUNTIME_BYTEORDER_H
#define WTRUNTIME_BYTEORDER_H

#include "config.h"
#include "wtruntime.h"

/* Type that holds a byte-order info. This type must be 1 byte. */
/* Only the value returned by wtr_byteorder() shall be used. */
typedef wt_uint8_t wt_bo;

struct wtr_swaptab
{
	wt_uint16_t (*const ld16)(wt_uint16_t);
	wt_uint16_t (*const st16)(wt_uint16_t);
	wt_uint32_t (*const ld32)(wt_uint32_t);
	wt_uint32_t (*const st32)(wt_uint32_t);
	wt_uint64_t (*const ld64)(wt_uint64_t);
	wt_uint64_t (*const st64)(wt_uint64_t);
};

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

wt_bo wtr_byteorder(void);

/* get the swaptable for the specified byteorder */
const struct wtr_swaptab* wtr_swaptab(wt_bo bo);

wt_uint16_t wtr_swap16(wt_uint16_t x);
wt_uint32_t wtr_swap32(wt_uint32_t x);
wt_uint64_t wtr_swap64(wt_uint64_t x);
wt_uint16_t wtr_botoh16(wt_uint16_t x, wt_bo bo);
wt_uint32_t wtr_botoh32(wt_uint32_t x, wt_bo bo);
wt_uint64_t wtr_botoh64(wt_uint64_t x, wt_bo bo);
wt_uint16_t wtr_htobo16(wt_uint16_t x, wt_bo bo);
wt_uint32_t wtr_htobo32(wt_uint32_t x, wt_bo bo);
wt_uint64_t wtr_htobo64(wt_uint64_t x, wt_bo bo);

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* WTRUNTIME_BYTEORDER_H */
