#include "config_private.h"
#include "convert.h"
#include <stdlib.h>
#if WTRUNTIME_ENABLE_FLOAT32 || WTRUNTIME_ENABLE_FLOAT64
#  include <stdio.h>
#endif /* WTRUNTIME_ENABLE_FLOAT32 || WTRUNTIME_ENABLE_FLOAT64 */

#define MOD10(_m_,_i_)          (_m_) = ((_i_) % 10)
#define DIV10(_i_)              (_i_) = ((_i_) / 10)
#define MUL10(_i_)              (_i_) = (((_i_)<<3) + ((_i_)<<1))
#define ADD_MUL10(_n_,_i_)      MUL10(_i_); (_i_) += (_n_)

/*================================================================< i32 >==*/

int wtr_convert_str_to_i32(union wtr_genvalue* const gv, const char* str)
{
	wt_int32_t value    = 0;
	wt_bool    negative = WT_FALSE;

	if('-'==str[0])
	{
		negative = WT_TRUE;
		++str;
	}

	while(*str)
	{
		char ch = *str++;

		switch(ch)
		{
		case '0': MUL10(value);       break;
		case '1': ADD_MUL10(1,value); break;
		case '2': ADD_MUL10(2,value); break;
		case '3': ADD_MUL10(3,value); break;
		case '4': ADD_MUL10(4,value); break;
		case '5': ADD_MUL10(5,value); break;
		case '6': ADD_MUL10(6,value); break;
		case '7': ADD_MUL10(7,value); break;
		case '8': ADD_MUL10(8,value); break;
		case '9': ADD_MUL10(9,value); break;
		default:                      break;
		}
	}

	if(negative)
	{
		gv->v_i32 = -value;
	}
	else
	{
		gv->v_i32 = value;
	}

	return 0;
}

char* wtr_convert_i32_to_str(union wtr_genvalue gv, char* begin, char* end)
{
	wt_int32_t value = gv.v_i32;

	wt_bool    negative = (value<0);
	wt_int32_t m;
	char*      p;

	if(0==value)
	{
		begin[0] = '0';
		begin[1] = '\0';
		return begin;
	}

	if(negative)
	{
		value = -value;
	}

	p = end - 1;
	*p = '\0';

	while(value && p>begin)
	{
		MOD10(m,value);
		*--p = (char)('0' + m);
		DIV10(value);
	}

	if(negative)
	{
		*--p = '-';
	}

	return p;
}

/*================================================================< u32 >==*/

int wtr_convert_str_to_u32(union wtr_genvalue* const gv, const char* str)
{
	wt_uint32_t value = 0;

	while(*str)
	{
		char ch = *str++;

		switch(ch)
		{
		case '0': MUL10(value);       break;
		case '1': ADD_MUL10(1,value); break;
		case '2': ADD_MUL10(2,value); break;
		case '3': ADD_MUL10(3,value); break;
		case '4': ADD_MUL10(4,value); break;
		case '5': ADD_MUL10(5,value); break;
		case '6': ADD_MUL10(6,value); break;
		case '7': ADD_MUL10(7,value); break;
		case '8': ADD_MUL10(8,value); break;
		case '9': ADD_MUL10(9,value); break;
		default:                      break;
		}
	}

	gv->v_u32 = value;

	return 0;
}

char* wtr_convert_u32_to_str(union wtr_genvalue gv, char* begin, char* end)
{
	wt_uint32_t value = gv.v_u32;

	wt_uint32_t m;
	char*       p;

	if(0==value)
	{
		begin[0] = '0';
		begin[1] = '\0';
		return begin;
	}

	p  = end - 1;
	*p = '\0';

	while(value && p>=begin)
	{
		MOD10(m,value);
		*--p = (char)('0' + m);
		DIV10(value);
	}

	return p;
}

/*================================================================< i64 >==*/

int wtr_convert_str_to_i64(union wtr_genvalue* const gv, const char* str)
{
	wt_int64_t value    = 0;
	wt_bool    negative = WT_FALSE;

	if('-'==str[0])
	{
		negative = WT_TRUE;
		++str;
	}

	while(*str)
	{
		char ch = *str++;

		switch(ch)
		{
		case '0': MUL10(value);       break;
		case '1': ADD_MUL10(1,value); break;
		case '2': ADD_MUL10(2,value); break;
		case '3': ADD_MUL10(3,value); break;
		case '4': ADD_MUL10(4,value); break;
		case '5': ADD_MUL10(5,value); break;
		case '6': ADD_MUL10(6,value); break;
		case '7': ADD_MUL10(7,value); break;
		case '8': ADD_MUL10(8,value); break;
		case '9': ADD_MUL10(9,value); break;
		default:                      break;
		}
	}

	if(negative)
	{
		gv->v_i64 = -value;
	}
	else
	{
		gv->v_i64 = value;
	}

	return 0;
}

char* wtr_convert_i64_to_str(union wtr_genvalue gv, char* begin, char* end)
{
	wt_int64_t value = gv.v_i64;

	wt_bool    negative = (value<0);
	wt_int64_t m;
	char*      p;

	if(0==value)
	{
		begin[0] = '0';
		begin[1] = '\0';
		return begin;
	}

	if(negative)
	{
		value = -value;
	}

	p = end - 1;
	*p = '\0';

	while(value && p>begin)
	{
		MOD10(m,value);
		*--p = (char)('0' + m);
		DIV10(value);
	}

	if(negative)
	{
		*--p = '-';
	}

	return p;
}

/*================================================================< u64 >==*/

int wtr_convert_str_to_u64(union wtr_genvalue* const gv, const char* str)
{
	wt_uint64_t value = 0;

	while(*str)
	{
		char ch = *str++;

		switch(ch)
		{
		case '0': MUL10(value);       break;
		case '1': ADD_MUL10(1,value); break;
		case '2': ADD_MUL10(2,value); break;
		case '3': ADD_MUL10(3,value); break;
		case '4': ADD_MUL10(4,value); break;
		case '5': ADD_MUL10(5,value); break;
		case '6': ADD_MUL10(6,value); break;
		case '7': ADD_MUL10(7,value); break;
		case '8': ADD_MUL10(8,value); break;
		case '9': ADD_MUL10(9,value); break;
		default:                      break;
		}
	}

	gv->v_u64 = value;

	return 0;
}

char* wtr_convert_u64_to_str(union wtr_genvalue gv, char* begin, char* end)
{
	wt_uint64_t value = gv.v_u64;

	wt_uint64_t m;
	char*       p;

	if(0==value)
	{
		begin[0] = '0';
		begin[1] = '\0';
		return begin;
	}

	p  = end - 1;
	*p = '\0';

	while(value && p>=begin)
	{
		MOD10(m,value);
		*--p = (char)('0' + m);
		DIV10(value);
	}

	return p;
}

/*================================================================< f32 >==*/

#if WTRUNTIME_ENABLE_FLOAT32

int wtr_convert_str_to_f32(union wtr_genvalue* const gv, const char* str)
{
	gv->v_f32 = (wt_float32_t)atof(str);
	return 0;
}

char* wtr_convert_f32_to_str(union wtr_genvalue gv, char* begin, char* end)
{
	/* TODO: deal with 'end' */
	((void)end);
	sprintf(begin, "%f", gv.v_f32);
	return begin;
}

#endif /* WTRUNTIME_ENABLE_FLOAT32 */

/*================================================================< f64 >==*/

#if WTRUNTIME_ENABLE_FLOAT64

int wtr_convert_str_to_f64(union wtr_genvalue* const gv, const char* str)
{
	gv->v_f64 = atof(str);
	return 0;
}

char* wtr_convert_f64_to_str(union wtr_genvalue gv, char* begin, char* end)
{
	/* TODO: deal with 'end' */
	((void)end);
	sprintf(begin, "%lf", gv.v_f64);
	return begin;
}

#endif /* WTRUNTIME_ENABLE_FLOAT64 */

/*================================================================< bool >==*/

#define MATCHANY_2(_val_,_const1_,_const2_) (((_const1_)==(_val_)) || ((_const2_)==(_val_)))

int wtr_convert_str_to_bool(union wtr_genvalue* const gv, const char* str)
{
	switch(*str)
	{
	case '1':
		gv->v_bool = !str[1];
		break;

	case 'o':
	case 'O':
		gv->v_bool = !str[2] && MATCHANY_2(str[1], 'n', 'N');
		break;

	case 'y':
	case 'Y':
		gv->v_bool = !str[3] && MATCHANY_2(str[1], 'e', 'E') &&
			MATCHANY_2(str[2], 's', 'S');
		break;

	case 't':
	case 'T':
		gv->v_bool = !str[4] && MATCHANY_2(str[1], 'r', 'R') &&
			MATCHANY_2(str[2], 'u', 'U') && MATCHANY_2(str[3], 'e', 'E');
		break;

	default:
		gv->v_bool = WT_FALSE;
	}

	return 0;
}

char* wtr_convert_bool_to_str(union wtr_genvalue gv, char* begin, char* end)
{
	char* ptr = begin;

	if(end-begin<6) { return NULL; }

	if(gv.v_bool)
	{
		*ptr++ = 't';
		*ptr++ = 'r';
		*ptr++ = 'u';
		*ptr++ = 'e';
	}
	else
	{
		*ptr++ = 'f';
		*ptr++ = 'a';
		*ptr++ = 'l';
		*ptr++ = 's';
		*ptr++ = 'e';
	}

	*ptr = '\0';

	return begin;
}

/*==========================================================================*/
