#include "config_private.h"
#include "lreader.h"
#include "vm.h"
#include <string.h>

static size_t clamp_offs(size_t size, wt_uint32_t offs, size_t n)
{
	if(size<=offs)
	{
		return 0;
	}

	size -= offs;

	if(size<n)
	{
		return size;
	}

	return n;
}

static size_t wtr_lr_sys_readsys(struct wtr_lreader* const lr, void* buf,
	wt_uint32_t offs, size_t n)
{
	n = clamp_offs(lr->size, offs, n);
	memcpy(buf, lr->u.sys.ptr+offs, n);
	return n;
}

static size_t wtr_lr_sys_readvma(struct wtr_lreader* const lr, wt_vmptr vma,
	wt_uint32_t offs, size_t n)
{
	n = clamp_offs(lr->size, offs, n);

	if(0!=wtr_vma_write(lr->vm, vma, lr->u.sys.ptr+offs, n))
	{
		return 0;
	}

	return n;
}

static wt_uint8_t wtr_lr_sys_read8(struct wtr_lreader* const lr,
	wt_uint32_t offs)
{
	wt_uint8_t x = 0;
	const size_t n = clamp_offs(lr->size, offs, sizeof(x));
	memcpy(&x, lr->u.sys.ptr+offs, n);
	return x;
}

static wt_uint16_t wtr_lr_sys_read16(struct wtr_lreader* const lr,
	wt_uint32_t offs)
{
	wt_uint16_t x = 0;
	const size_t n = clamp_offs(lr->size, offs, sizeof(x));
	memcpy(&x, lr->u.sys.ptr+offs, n);
	return x;
}

static wt_uint32_t wtr_lr_sys_read32(struct wtr_lreader* const lr,
	wt_uint32_t offs)
{
	wt_uint32_t x = 0;
	const size_t n = clamp_offs(lr->size, offs, sizeof(x));
	memcpy(&x, lr->u.sys.ptr+offs, n);
	return x;
}

static wt_uint64_t wtr_lr_sys_read64(struct wtr_lreader* const lr,
	wt_uint32_t offs)
{
	wt_uint64_t x = 0;
	const size_t n = clamp_offs(lr->size, offs, sizeof(x));
	memcpy(&x, lr->u.sys.ptr+offs, n);
	return x;
}

static wt_uint16_t wtr_lr_sys_read16_swap(struct wtr_lreader* const lr,
	wt_uint32_t offs)
{
	return wtr_swap16(wtr_lr_sys_read16(lr, offs));
}

static wt_uint32_t wtr_lr_sys_read32_swap(struct wtr_lreader* const lr,
	wt_uint32_t offs)
{
	return wtr_swap32(wtr_lr_sys_read32(lr, offs));
}

static wt_uint64_t wtr_lr_sys_read64_swap(struct wtr_lreader* const lr,
	wt_uint32_t offs)
{
	return wtr_swap64(wtr_lr_sys_read64(lr, offs));
}

static size_t wtr_lr_vm_readsys(struct wtr_lreader* const lr, void* buf,
	wt_uint32_t offs, size_t n)
{
	n = clamp_offs(lr->size, offs, n);

	if(0!=wtr_vma_read(lr->vm, lr->u.vm.vma+offs, buf, n))
	{
		return 0;
	}

	return n;
}

static size_t wtr_lr_vm_readvma(struct wtr_lreader* const lr, wt_vmptr vma,
	wt_uint32_t offs, size_t n)
{
	n = clamp_offs(lr->size, offs, n);

	if(0!=wtr_vma_memcpy(lr->vm, vma, lr->u.vm.vma+offs, n))
	{
		return 0;
	}

	return n;
}

static wt_uint8_t wtr_lr_vm_read8(struct wtr_lreader* const lr,
	wt_uint32_t offs)
{
	wt_uint8_t x = 0;
	const size_t n = clamp_offs(lr->size, offs, sizeof(x));
	wtr_vma_read(lr->vm, lr->u.vm.vma+offs, &x, n);
	return x;
}

static wt_uint16_t wtr_lr_vm_read16(struct wtr_lreader* const lr,
	wt_uint32_t offs)
{
	wt_uint16_t x = 0;
	const size_t n = clamp_offs(lr->size, offs, sizeof(x));
	wtr_vma_read(lr->vm, lr->u.vm.vma+offs, &x, n);
	return x;
}

static wt_uint32_t wtr_lr_vm_read32(struct wtr_lreader* const lr,
	wt_uint32_t offs)
{
	wt_uint32_t x = 0;
	const size_t n = clamp_offs(lr->size, offs, sizeof(x));
	wtr_vma_read(lr->vm, lr->u.vm.vma+offs, &x, n);
	return x;
}

static wt_uint64_t wtr_lr_vm_read64(struct wtr_lreader* const lr,
	wt_uint32_t offs)
{
	wt_uint64_t x = 0;
	const size_t n = clamp_offs(lr->size, offs, sizeof(x));
	wtr_vma_read(lr->vm, lr->u.vm.vma+offs, &x, n);
	return x;
}

static wt_uint16_t wtr_lr_vm_read16_swap(struct wtr_lreader* const lr,
	wt_uint32_t offs)
{
	return wtr_swap16(wtr_lr_vm_read16(lr, offs));
}

static wt_uint32_t wtr_lr_vm_read32_swap(struct wtr_lreader* const lr,
	wt_uint32_t offs)
{
	return wtr_swap32(wtr_lr_vm_read32(lr, offs));
}

static wt_uint64_t wtr_lr_vm_read64_swap(struct wtr_lreader* const lr,
	wt_uint32_t offs)
{
	return wtr_swap64(wtr_lr_vm_read64(lr, offs));
}

void wtr_lreader_init_sys(struct wtr_lreader* const lr,
	struct wtr_vm* const vm, const void* const data, size_t size,
	const wt_bo bo)
{
	lr->read_sys = wtr_lr_sys_readsys;
	lr->read_vma = wtr_lr_sys_readvma;

	if(bo==wtr_byteorder())
	{
		lr->read8  = wtr_lr_sys_read8;
		lr->read16 = wtr_lr_sys_read16;
		lr->read32 = wtr_lr_sys_read32;
		lr->read64 = wtr_lr_sys_read64;
	}
	else
	{
		lr->read8  = wtr_lr_sys_read8;
		lr->read16 = wtr_lr_sys_read16_swap;
		lr->read32 = wtr_lr_sys_read32_swap;
		lr->read64 = wtr_lr_sys_read64_swap;
	}

	lr->vm        = vm;
	lr->size      = size;
	lr->u.sys.ptr = (const wt_uint8_t*)data;
}

void wtr_lreader_init_vm(struct wtr_lreader* const lr,
	struct wtr_vm* const vm, wt_vmptr data, size_t size, const wt_bo bo)
{
	lr->read_sys = wtr_lr_vm_readsys;
	lr->read_vma = wtr_lr_vm_readvma;

	if(bo==wtr_byteorder())
	{
		lr->read8  = wtr_lr_vm_read8;
		lr->read16 = wtr_lr_vm_read16;
		lr->read32 = wtr_lr_vm_read32;
		lr->read64 = wtr_lr_vm_read64;
	}
	else
	{
		lr->read8  = wtr_lr_vm_read8;
		lr->read16 = wtr_lr_vm_read16_swap;
		lr->read32 = wtr_lr_vm_read32_swap;
		lr->read64 = wtr_lr_vm_read64_swap;
	}

	lr->vm       = vm;
	lr->size     = size;
	lr->u.vm.vma = data;
}

wt_uint32_t wtr_lreader_strlen(struct wtr_lreader* const lr,
	const wt_uint32_t start_offs)
{
	wt_uint32_t offs = start_offs;

	while(0!=lr->read8(lr, offs))
	{
		++offs;
	}

	return offs - start_offs;
}
