#ifndef WTRUNTIME_LREADER_H
#define WTRUNTIME_LREADER_H

#include "config.h"
#include "types.h"
#include "byteorder.h"
#include "wtruntime.h"

struct wtr_vm;

struct wtr_lreader
{
	wt_uint8_t  (*read8) (struct wtr_lreader* const lr, wt_uint32_t offs);
	wt_uint16_t (*read16)(struct wtr_lreader* const lr, wt_uint32_t offs);
	wt_uint32_t (*read32)(struct wtr_lreader* const lr, wt_uint32_t offs);
	wt_uint64_t (*read64)(struct wtr_lreader* const lr, wt_uint32_t offs);

	/* reads data to system memory as is, no byteorder swap */
	size_t (*read_sys)(struct wtr_lreader* const lr, void* buf,
		wt_uint32_t offs, size_t n);

	/* reads data to virtual machine memory as is, no byteorder swap */
	size_t (*read_vma)(struct wtr_lreader* const lr, wt_vmptr vma,
		wt_uint32_t offs, size_t n);

	struct wtr_vm* vm;
	size_t         size;

	union
	{
	struct
	{
	wt_vmptr       vma;
	} vm;
	struct
	{
	const wt_uint8_t* ptr;
	} sys;
	} u;
};

void wtr_lreader_init_sys(struct wtr_lreader* const lr,
	struct wtr_vm* const vm, const void* const data, size_t size,
	const wt_bo bo);

void wtr_lreader_init_vm(struct wtr_lreader* const lr,
	struct wtr_vm* const vm, wt_vmptr data, size_t size, const wt_bo bo);

wt_uint32_t wtr_lreader_strlen(struct wtr_lreader* const lr,
	const wt_uint32_t start_offs);

#endif /* WTRUNTIME_LREADER_H */
